@@ -1,22 +1,37 @@
-import collections
+try:
+    from collections.abc import Mapping, MutableMapping
+except:
+    from collections import Mapping, MutableMapping
 
 import six
 from wtforms import Form
+try:
+    from wtforms_sqlalchemy.fields import (
+        QuerySelectField,
+        QuerySelectMultipleField
+    )
+    HAS_SQLALCHEMY_SUPPORT = True
+except ImportError:
+    try:
+        from wtforms.ext.sqlalchemy.fields import (
+            QuerySelectField,
+            QuerySelectMultipleField
+        )
+        HAS_SQLALCHEMY_SUPPORT = True
+    except ImportError:
+        HAS_SQLALCHEMY_SUPPORT = False
 from wtforms.fields import (
+    _unset_value,
     BooleanField,
     Field,
     FieldList,
+    FileField,
     FormField,
-    TextField,
-    _unset_value
+    StringField
 )
-from wtforms.ext.sqlalchemy.fields import (
-    QuerySelectField, QuerySelectMultipleField
-)
-from wtforms.validators import Optional, DataRequired
+from wtforms.validators import DataRequired, Optional
 
-
-__version__ = '0.2.8'
+__version__ = '0.3.4'
 
 
 class InvalidData(Exception):
@@ -45,7 +60,7 @@ def flatten_json(
         >>> flatten_json(MyForm, {'a': {'b': 'c'}})
         {'a-b': 'c'}
     """
-    if not isinstance(json, collections.Mapping):
+    if not isinstance(json, Mapping):
         raise InvalidData(
             u'This function only accepts dict-like data structures.'
         )
@@ -69,19 +84,24 @@ def flatten_json(
                 raise InvalidData(u"Key '%s' is not valid field class." % key)
 
         new_key = parent_key + separator + key if parent_key else key
-        if isinstance(value, collections.MutableMapping):
+        if isinstance(value, MutableMapping):
             if issubclass(field_class, FormField):
+                nested_form_class = unbound_field.bind(Form(), '').form_class
                 items.extend(
-                    flatten_json(unbound_field.args[0], value, new_key)
+                    flatten_json(nested_form_class, value, new_key)
                     .items()
                 )
             else:
                 items.append((new_key, value))
         elif isinstance(value, list):
             if issubclass(field_class, FieldList):
+                nested_unbound_field = unbound_field.bind(
+                    Form(),
+                    ''
+                ).unbound_field
                 items.extend(
                     flatten_json_list(
-                        unbound_field.args[0],
+                        nested_unbound_field,
                         value,
                         new_key,
                         separator
@@ -98,11 +118,16 @@ def flatten_json_list(field, json, parent_key='', separator='-'):
     items = []
     for i, item in enumerate(json):
         new_key = parent_key + separator + str(i)
-        if isinstance(item, list):
-            items.extend(flatten_json_list(item, new_key, separator))
-        elif isinstance(item, dict):
+        if (
+            isinstance(item, dict) and
+            issubclass(getattr(field, 'field_class'), FormField)
+        ):
+            nested_class = field.field_class(
+                *field.args,
+                **field.kwargs
+            ).bind(Form(), '').form_class
             items.extend(
-                flatten_json(field.args[0], item, new_key, separator)
+                flatten_json(nested_class, item, new_key)
                 .items()
             )
         else:
@@ -152,18 +177,16 @@ def monkey_patch_field_process(func):
         call_original_func = True
         if not isinstance(self, FormField):
 
-            self.is_missing = True
-            if formdata:
-                if self.name in formdata:
-                    if (
-                        len(formdata.getlist(self.name)) == 1 and
-                        formdata.getlist(self.name) == [None]
-                    ):
-                        call_original_func = False
-                        self.data = None
-                    self.is_missing = not bool(formdata.getlist(self.name))
-                else:
-                    self.is_missing = True
+            if formdata and self.name in formdata:
+                if (
+                    len(formdata.getlist(self.name)) == 1 and
+                    formdata.getlist(self.name) == [None]
+                ):
+                    call_original_func = False
+                    self.data = None
+                self.is_missing = not bool(formdata.getlist(self.name))
+            else:
+                self.is_missing = True
 
         if call_original_func:
             func(self, formdata, data=data)
@@ -176,9 +199,12 @@ def monkey_patch_field_process(func):
             self.form._is_missing = False
             self.form._patch_data = None
 
-        if isinstance(self, TextField):
-            if self.data is None:
-                self.data = u''
+        if isinstance(self, StringField) and not isinstance(self, FileField):
+            if not self.data:
+                try:
+                    self.data = self.default()
+                except TypeError:
+                    self.data = self.default
             else:
                 self.data = six.text_type(self.data)
 
@@ -241,7 +267,7 @@ def field_list_is_missing(self):
 
 def monkey_patch_process_formdata(func):
     def process_formdata(self, valuelist):
-        valuelist = map(six.text_type, valuelist)
+        valuelist = list(map(six.text_type, valuelist))
 
         return func(self, valuelist)
     return process_formdata
@@ -253,12 +279,14 @@ def init():
     Form.from_json = from_json
     Form.patch_data = patch_data
     FieldList.patch_data = patch_data
-    QuerySelectField.process_formdata = monkey_patch_process_formdata(
-        QuerySelectField.process_formdata
-    )
-    QuerySelectMultipleField.process_formdata = monkey_patch_process_formdata(
-        QuerySelectMultipleField.process_formdata
-    )
+    if HAS_SQLALCHEMY_SUPPORT:
+        QuerySelectField.process_formdata = monkey_patch_process_formdata(
+            QuerySelectField.process_formdata
+        )
+        QuerySelectMultipleField.process_formdata = \
+            monkey_patch_process_formdata(
+                QuerySelectMultipleField.process_formdata
+            )
     Field.process = monkey_patch_field_process(Field.process)
     FormField.process = monkey_patch_field_process(FormField.process)
     BooleanField.false_values = BooleanField.false_values + (False,)
